import logging
from typing import Any, override

import transformers
from rich.console import Console
from rich.logging import RichHandler
from rich.progress import (
    BarColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Table
from torch.utils.data import DataLoader

console = Console()
rich_handler = RichHandler(console=console)
logger: logging.Logger = logging.getLogger(name=__name__)


def has_length(dataloader: DataLoader) -> bool:
    try:
        len(dataloader)
        return True
    except (TypeError, ValueError, AttributeError):
        return False


class ProgressCallback(transformers.TrainerCallback):
    def __init__(self, max_str_len: int = 100):
        """
        Initialize the callback with optional max_str_len parameter to control string truncation length.

        Args:
            max_str_len (`int`):
                Maximum length of strings to display in logs.
                Longer strings will be truncated with a message.
        """
        self.max_str_len: int = max_str_len
        self.console: Console = Console()
        self.current_step: int = 0
        self.training_progress: Progress | None = None
        self.train_task_id: Any = None
        self.prediction_progress: Progress | None = None
        self.eval_task_id: Any = None

    @override
    def on_train_begin(self, args, state, control, **kwargs):  # pyright: ignore
        if state.is_world_process_zero:
            self.training_progress = Progress(
                TextColumn("[bold blue]{task.description}"),
                BarColumn(bar_width=None),
                TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
                console=self.console,
                expand=True,
            )
            self.train_task_id = self.training_progress.add_task(
                "[green]Training", total=state.max_steps
            )
            self.training_progress.start()
        self.current_step = 0

    @override
    def on_step_end(self, args, state, control, **kwargs):  # pyright: ignore
        if state.is_world_process_zero and self.training_progress is not None:
            steps_completed = state.global_step - self.current_step
            self.training_progress.update(self.train_task_id, advance=steps_completed)
            self.current_step = state.global_step

    @override
    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):  # pyright: ignore
        if state.is_world_process_zero and has_length(eval_dataloader):
            if self.prediction_progress is None:
                self.prediction_progress = Progress(
                    TextColumn("[bold yellow]{task.description}"),
                    BarColumn(bar_width=None),
                    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                    console=self.console,
                    expand=True,
                )
                self.eval_task_id = self.prediction_progress.add_task(
                    "[yellow]Evaluating", total=len(eval_dataloader)
                )
                self.prediction_progress.start()
            else:
                self.prediction_progress.update(self.eval_task_id, advance=1)

    @override
    def on_evaluate(self, args, state, control, **kwargs):  # pyright: ignore
        if state.is_world_process_zero and self.prediction_progress is not None:
            self.prediction_progress.stop()
            self.prediction_progress = None

    @override
    def on_predict(self, args, state, control, **kwargs):  # pyright: ignore
        if state.is_world_process_zero and self.prediction_progress is not None:
            self.prediction_progress.stop()
            self.prediction_progress = None

    @override
    def on_log(self, args, state, control, logs=None, **kwargs):  # pyright: ignore
        if state.is_world_process_zero:
            shallow_logs = {}
            for k, v in logs.items():  # pyright: ignore
                if isinstance(v, str) and len(v) > self.max_str_len:
                    shallow_logs[k] = (
                        f"[String too long to display, length: {len(v)} > {self.max_str_len}. "
                        "Consider increasing `max_str_len` if needed.]"
                    )
                elif isinstance(v, float):
                    if abs(v) < 0.0001 and v != 0.0:
                        shallow_logs[k] = f"{v:.8f}"
                    else:
                        shallow_logs[k] = f"{v:.4f}"
                else:
                    shallow_logs[k] = v
            _ = shallow_logs.pop("total_flos", None)

            if "epoch" in shallow_logs and isinstance(shallow_logs["epoch"], float):
                shallow_logs["epoch"] = f"{shallow_logs['epoch']:.4f}"

            table = Table(show_header=True, header_style="bold magenta")
            table.add_column("Metric")
            table.add_column("Value")

            for key, value in shallow_logs.items():
                table.add_row(key, str(value))

            if self.training_progress is not None:
                self.training_progress.stop()
                self.console.print(table)
                self.training_progress.start()
            else:
                self.console.print(table)

    @override
    def on_train_end(self, args, state, control, **kwargs):  # pyright: ignore
        if state.is_world_process_zero:
            if self.training_progress is not None:
                self.training_progress.stop()
                self.training_progress = None
            self.console.print("[bold green]Training completed![/bold green]")



import logging
from abc import ABC, abstractmethod
from typing import NamedTuple, override

import torch.nn as nn
from torch import Tensor, device, dtype

from ...config.config_schema import ConnectorConfig

log: logging.Logger = logging.getLogger(name=__name__)


class ProcessedVisualFeatures(NamedTuple):
    features: list[Tensor]


class ForwardOutput(NamedTuple):
    embeddings: Tensor
    attention_mask: Tensor


class Connector(nn.Module, ABC):
    def __init__(
        self,
        config: ConnectorConfig,
        image_hidden_size: int,
        text_hidden_size: int,
        torch_dtype: dtype,
        torch_device: device,
    ) -> None:
        super().__init__()
        self.config: ConnectorConfig = config
        self.name: str = self.config.name
        self.torch_dtype: dtype = torch_dtype
        self.torch_device: device = torch_device

        self.image_hidden_size: int = image_hidden_size
        self.text_hidden_size: int = text_hidden_size
        self.projection_layer: nn.Module = self.build_projection_layer().to(dtype=self.torch_dtype)
        self.initialize_layers()

    @abstractmethod
    def _build_projection_layer(self) -> nn.Module:
        pass

    def build_projection_layer(self) -> nn.Module:
        return self._build_projection_layer()

    @abstractmethod
    def _initialize_layers(self) -> None:
        pass

    def initialize_layers(self) -> None:
        self._initialize_layers()

    @abstractmethod
    def projection(self, visual_features: Tensor) -> Tensor:
        pass

    @override
    def forward(self, visual_features: Tensor):
        return self.projection(visual_features)



import os
import sys

from .vlm import main

if __name__ == "__main__":
    local_rank = None
    i = 0
    while i < len(sys.argv):
        if sys.argv[i].startswith("--local_rank="):
            local_rank = int(sys.argv[i].split("=")[1])
            sys.argv.pop(i)
        else:
            i += 1

    if local_rank is not None:
        os.environ["LOCAL_RANK"] = str(local_rank)
    main()



import logging
from typing import Any, cast, override

import torch.nn as nn
from torch import FloatTensor, LongTensor, device, dtype
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from transformers.utils import ModelOutput

from ...config.config_schema import LLMConfig
from .base import LanguageModel

log: logging.Logger = logging.getLogger(name=__name__)


class HFLLMLanguageModel(LanguageModel):
    def __init__(
        self, config: LLMConfig, torch_dtype: dtype, torch_device: device, attn_implementation: str
    ) -> None:
        super().__init__(config, torch_dtype, torch_device, attn_implementation)

    @override
    def get_embedding_layer(self) -> nn.Module:
        return self.language_model.get_input_embeddings()

    @override
    def _build_tokenizer(self) -> PreTrainedTokenizer:
        return cast(
            PreTrainedTokenizer,
            AutoTokenizer.from_pretrained(
                self.hf_name,
                trust_remote_code=True,
                model_max_length=self.config.max_seq_length,
                use_fast=True,
            ),
        )

    @override
    def _build_language_model(self) -> PreTrainedModel:
        llm = cast(
            PreTrainedModel,
            AutoModelForCausalLM.from_pretrained(
                self.hf_name,
                low_cpu_mem_usage=True,
                trust_remote_code=True,
                attn_implementation=self.attn_implementation,
                torch_dtype=self.torch_dtype,
            ),
        )
        if llm.device == device("meta"):
            llm.to_empty(device=self.torch_device)
        else:
            llm.to(device=self.torch_device)
        return llm

    @override
    def _build_hf_config(self) -> PretrainedConfig:
        return cast(
            PretrainedConfig, AutoConfig.from_pretrained(self.hf_name, trust_remote_code=True)
        )

    @override
    def generate(
        self,
        position_ids: LongTensor | None = None,
        attention_mask: LongTensor | None = None,
        inputs_embeds: FloatTensor | None = None,
        **kwargs: Any,
    ) -> ModelOutput | LongTensor:
        return self.language_model.generate(
            position_ids=position_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            **kwargs,
        )



import logging
from typing import Any

import torch.nn as nn
from torch.nn.parameter import Parameter
from transformers import PreTrainedModel
from transformers.pytorch_utils import ALL_LAYERNORM_LAYERS
from transformers.trainer_pt_utils import get_parameter_names

from .training_arguments import TrainingArguments

log = logging.getLogger(__name__)


def configure_optimizers(model: PreTrainedModel | nn.Module, trainer_config: TrainingArguments):
    log.info("configure_optimizers")
    decay_parameters = get_parameter_names(model, ALL_LAYERNORM_LAYERS)
    decay_parameters = [name for name in decay_parameters if "bias" not in name]

    param_groups: dict[str, dict[str, list[Parameter]]] = {}

    visual_encoder_params = collect_param_groups(model.visual_encoder, decay_parameters)
    param_groups["visual_encoder"] = visual_encoder_params

    language_model_params = collect_param_groups(model.language_model, decay_parameters)
    param_groups["language_model"] = language_model_params

    connector_params = collect_param_groups(model.connector, decay_parameters)
    param_groups["connector"] = connector_params

    return build_optimizer_params(trainer_config, param_groups)


def collect_param_groups(
    module: PreTrainedModel | nn.Module, decay_parameters: list[str]
) -> dict[str, list[Parameter]]:
    decay_params = [
        p for n, p in module.named_parameters() if p.requires_grad and n in decay_parameters
    ]

    no_decay_params = [
        p for n, p in module.named_parameters() if p.requires_grad and n not in decay_parameters
    ]

    if decay_params or no_decay_params:
        return {"decay": decay_params, "no_decay": no_decay_params}
    return {"decay": [], "no_decay": []}


def build_optimizer_params(
    trainer_config: TrainingArguments,
    param_groups: dict[str, dict[str, list[Parameter]]],
) -> list[dict[str, Any]]:
    optimizer_params: list[dict[str, Any]] = []

    optimizer_params.extend(
        get_module_param_groups(
            module_name="visual_encoder",
            param_groups=param_groups,
            weight_decay=trainer_config.visual_encoder_wd,
            learning_rate=trainer_config.visual_encoder_lr,
        )
    )

    optimizer_params.extend(
        get_module_param_groups(
            module_name="language_model",
            param_groups=param_groups,
            weight_decay=trainer_config.language_model_wd,
            learning_rate=trainer_config.language_model_lr,
        )
    )

    optimizer_params.extend(
        get_module_param_groups(
            module_name="connector",
            param_groups=param_groups,
            weight_decay=trainer_config.connector_wd,
            learning_rate=trainer_config.connector_lr,
        )
    )
    return optimizer_params


def get_module_param_groups(
    module_name: str,
    param_groups: dict[str, dict[str, list[Parameter]]],
    weight_decay: float,
    learning_rate: float,
) -> list[dict[str, Any]]:
    log.info(f"{module_name} lr: {learning_rate}, weight_decay: {weight_decay}")
    return [
        {
            "params": param_groups[module_name]["decay"],
            "weight_decay": weight_decay,
            "lr": learning_rate,
        },
        {
            "params": param_groups[module_name]["no_decay"],
            "weight_decay": 0.0,
            "lr": learning_rate,
        },
    ]



import logging
from collections.abc import Callable
from typing import Any, override

import torch
from omegaconf import OmegaConf
from torch import FloatTensor, LongTensor, Tensor, device, dtype
from transformers import (
    AutoConfig,
    AutoModel,
    GenerationConfig,
    GenerationMixin,
    LogitsProcessorList,
    PretrainedConfig,
    PreTrainedModel,
    StoppingCriteriaList,
)

from ..config import (
    ConnectorConfig,
    LLMConfig,
    ModelConfig,
    VisualEncoderConfig,
)
from .connectors import Connector, connector_map
from .language_models import LanguageModel
from .visual_encoders import VisualEncoder

log: logging.Logger = logging.getLogger(name=__name__)


class VLMConfig(PretrainedConfig):
    model_type: str = "vlm"

    def __init__(
        self,
        model_config_dict: dict[str, Any] = None,
        model_dtype: str = "float16",
        hidden_size: int | None = None,
        attn_implementation: str = "eager",
        torch_device: device = "cuda",
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.model_config_dict: dict[str, Any] = model_config_dict
        self.model_dtype: str = model_dtype
        self.hidden_size: int | None = hidden_size
        self.attn_implementation: str = attn_implementation
        self.torch_device: device = torch_device


class VLM(PreTrainedModel, GenerationMixin):
    config_class = VLMConfig  # pyright: ignore

    def __init__(
        self,
        config: VLMConfig,
    ) -> None:
        super().__init__(config)
        self.model_config: ModelConfig = self._process_config(config.model_config_dict)

        self.torch_dtype: dtype = getattr(torch, config.model_dtype)
        self.torch_device: torch.device = config.torch_device
        self.attn_implementation: str = config.attn_implementation

        # initialize components
        self.initialize_components()

    def _process_config(self, config: Any) -> Any:
        if isinstance(config, dict):
            return OmegaConf.create(config)
        return config

    # initialize all components
    def initialize_components(self) -> None:
        self._visual_encoder: VisualEncoder = self._build_visual_encoder()
        self._language_model: LanguageModel = self._build_language_model()
        self._connector: Connector = self._build_connector()

    @property
    def visual_encoder(self) -> VisualEncoder:
        return self._visual_encoder

    @property
    def language_model(self) -> LanguageModel:
        return self._language_model

    @property
    def connector(self) -> Connector:
        return self._connector

    @property
    @override
    def supports_gradient_checkpointing(self):
        return self.language_model.language_model.supports_gradient_checkpointing

    def _build_visual_encoder(self) -> VisualEncoder:
        encoder_config: VisualEncoderConfig = self.model_config.visual_encoder
        if encoder_config.type == "hf_visual_encoder":
            from .visual_encoders import HFVisualEncoder

            return HFVisualEncoder(encoder_config, self.torch_dtype, self.torch_device)
        else:
            error_msg = f"Unknown visual encoder type: {encoder_config.type}"
            log.error(error_msg)
            raise ValueError(error_msg)

    def _build_language_model(self) -> LanguageModel:
        llm_config: LLMConfig = self.model_config.llm
        if llm_config.type == "hf_llm":
            from .language_models import HFLLMLanguageModel

            return HFLLMLanguageModel(
                llm_config, self.torch_dtype, self.torch_device, self.attn_implementation
            )
        else:
            error_msg = f"Unknown language model type: {llm_config.type}"
            log.error(error_msg)
            raise ValueError(error_msg)

    def _build_connector(self) -> Connector:
        connector_config: ConnectorConfig = self.model_config.connector
        connector_class = connector_map.get(connector_config.type)
        if not connector_class:
            raise ValueError(f"Unsupported connector type: {connector_config.type}")
        return connector_class(
            connector_config,
            self.visual_encoder.hidden_size,
            self.language_model.hidden_size,
            self.torch_dtype,
            self.torch_device,
        )

    @override
    def forward(
        self,
        input_ids: Tensor | None = None,
        inputs_embeds: Tensor | None = None,
        attention_mask: Tensor | None = None,
        position_ids: LongTensor | None = None,
        past_key_values: list[FloatTensor] | None = None,
        labels: LongTensor | None = None,
        use_cache: bool | None = None,
        output_attentions: bool | None = None,
        output_hidden_states: bool | None = None,
        images: FloatTensor | None = None,
        image_sizes: list[list[int]] | None = None,
        return_dict: bool | None = None,
    ) -> torch.Tensor:
        if inputs_embeds is None:
            (input_ids, position_ids, attention_mask, past_key_values, inputs_embeds, labels) = (
                self.prepare_inputs_labels_for_multimodal(
                    input_ids,
                    position_ids,
                    attention_mask,
                    past_key_values,
                    labels,
                    images,
                )
            )
        return self.language_model(
            input_ids=input_ids,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            labels=labels,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

    @override
    def generate(
        self,
        inputs: Tensor | None = None,
        generation_config: GenerationConfig | None = None,
        logits_processor: LogitsProcessorList | None = None,
        stopping_criteria: StoppingCriteriaList | None = None,
        prefix_allowed_tokens_fn: Callable | None = None,
        synced_gpus: bool | None = None,
        assistant_model: PreTrainedModel | None = None,
        streamer: Any | None = None,
        negative_prompt_ids: Tensor | None = None,
        negative_prompt_attention_mask: Tensor | None = None,
        use_model_defaults: bool | None = None,
        images: FloatTensor | None = None,
        image_sizes: list[list[int]] | None = None,
        **kwargs: Any,
    ) -> Any:
        position_ids = kwargs.pop("position_ids", None)
        attention_mask = kwargs.pop("attention_mask", None)
        if "inputs_embeds" in kwargs:
            raise NotImplementedError("`inputs_embeds` is not supported")

        if images is not None:
            (inputs, position_ids, attention_mask, _, inputs_embeds, _) = (
                self.prepare_inputs_labels_for_multimodal(
                    inputs,
                    position_ids,
                    attention_mask,
                    None,
                    None,
                    images,
                    image_sizes=image_sizes,
                )
            )
        else:
            inputs_embeds = self.language_model.get_embedding_layer()(inputs)

        return self.language_model.generate(
            position_ids=position_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            **kwargs,
        )

    def freeze_visual_encoder(self, freeze: bool = True) -> None:
        for param in self.visual_encoder.visual_encoder.parameters():
            param.requires_grad = not freeze

    def freeze_language_model(self, freeze: bool = True, except_layer_norm: bool = False) -> None:
        for name, param in self.language_model.language_model.named_parameters():
            if except_layer_norm and (
                "layernorm" in name.lower() or "layer_norm" in name.lower() or "ln_" in name.lower()
            ):
                param.requires_grad = True
            else:
                param.requires_grad = not freeze
        for param in self.language_model.get_embedding_layer().parameters():
            param.requires_grad = False
        if self.model_config.llm.use_start_end_tokens:
            for param in self.language_model.get_embedding_layer().parameters():
                param.requires_grad = True

    def freeze_connector(self, freeze: bool = True) -> None:
        for param in self.connector.parameters():
            param.requires_grad = not freeze

    def set_trainable_params(self, config: dict[str, bool]) -> None:
        self.freeze_visual_encoder(not config.train_visual_encoder)
        self.freeze_language_model(not config.train_language_model)
        self.freeze_connector(not config.train_connector)
        self._log_trainable_params()

    def _log_trainable_params(self) -> None:
        trainable_params = sum(p.numel() for p in self.parameters() if p.requires_grad)
        total_params = sum(p.numel() for p in self.parameters())
        if total_params > 0:
            log.info(
                f"Trainable parameters: {trainable_params:,} ({trainable_params / total_params:.2%} of total)"
            )
        for module_name, module in [
            ("visual_encoder", self.visual_encoder),
            ("language_model", self.language_model),
            ("connector", self.connector),
        ]:
            trainable = sum(p.numel() for p in module.parameters() if p.requires_grad)
            total = sum(p.numel() for p in module.parameters())
            if total > 0:
                log.info(f"  - {module_name}: {trainable:,} ({trainable / total:.2%} of {total:,})")

    def encode_images(self, images: Tensor) -> tuple[Tensor, ...]:
        image_features = self.visual_encoder(images)
        image_features = self.connector(image_features)
        return image_features

    def unpad_image(self, tensor: Tensor, original_size: tuple[int, int]) -> Tensor:
        """
        Unpads a PyTorch tensor of a padded and resized image.

        Args:
        tensor (torch.Tensor): The image tensor, assumed to be in CxHxW format.
        original_size (tuple): The original size of PIL image (width, height).

        Returns:
        torch.Tensor: The unpadded image tensor.
        """
        original_width, original_height = original_size
        current_height, current_width = tensor.shape[1:]

        original_aspect_ratio = original_width / original_height
        current_aspect_ratio = current_width / current_height

        if original_aspect_ratio > current_aspect_ratio:
            scale_factor = current_width / original_width
            new_height = int(original_height * scale_factor)
            padding = (current_height - new_height) // 2
            unpadded_tensor = tensor[:, padding : current_height - padding, :]
        else:
            scale_factor = current_height / original_height
            new_width = int(original_width * scale_factor)
            padding = (current_width - new_width) // 2
            unpadded_tensor = tensor[:, :, padding : current_width - padding]

        return unpadded_tensor

    def prepare_inputs_labels_for_multimodal(
        self,
        input_ids: Tensor | None = None,
        position_ids: LongTensor | None = None,
        attention_mask: Tensor | None = None,
        past_key_values: list[FloatTensor] | None = None,
        labels: LongTensor | None = None,
        images: FloatTensor | None = None,
    ) -> tuple[
        Tensor | None,
        LongTensor | None,
        Tensor | None,
        list[FloatTensor] | None,
        Tensor | None,
        LongTensor | None,
    ]:
        visual_encoder = self.visual_encoder
        if visual_encoder is None or images is None or input_ids.shape[1] == 1:  # pyright: ignore
            return input_ids, position_ids, attention_mask, past_key_values, None, labels

        if isinstance(images, list) or images.ndim == 5:
            if isinstance(images, list):
                images = [x.unsqueeze(0) if x.ndim == 3 else x for x in images]  # pyright: ignore
            concat_images = torch.cat([image for image in images], dim=0)  # pyright: ignore
            image_features = self.encode_images(concat_images)
            split_sizes = [image.shape[0] for image in images]  # pyright: ignore
            image_features: tuple[Tensor, ...] = torch.split(image_features, split_sizes, dim=0)  # pyright: ignore
            image_features = [x.flatten(0, 1) for x in image_features]  # pyright: ignore
        else:
            image_features = self.encode_images(images)

        # Let's just add dummy tensors if they do not exist,
        # it is a headache to deal with None all the time.
        # But it is not ideal, and if you have a better idea,
        # please open an issue / submit a PR, thanks.
        _labels = labels
        _position_ids = position_ids
        _attention_mask = attention_mask
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids, dtype=torch.bool)
        else:
            attention_mask = attention_mask.bool()
        if position_ids is None:
            position_ids = torch.arange(
                0,
                input_ids.shape[1],  # pyright: ignore
                dtype=torch.long,
                device=input_ids.device,  # pyright: ignore
            )
        if labels is None:
            labels = torch.full_like(input_ids, self.model_config.llm.ignore_index)  # pyright: ignore

        # remove the padding using attention_mask -- FIXME
        _input_ids = input_ids
        input_ids = [
            cur_input_ids[cur_attention_mask]
            for cur_input_ids, cur_attention_mask in zip(input_ids, attention_mask, strict=False)
        ]  # pyright: ignore
        labels = [
            cur_labels[cur_attention_mask]
            for cur_labels, cur_attention_mask in zip(labels, attention_mask, strict=False)
        ]  # pyright: ignore

        new_input_embeds = []
        new_labels = []
        cur_image_idx = 0
        for batch_idx, cur_input_ids in enumerate(input_ids):
            num_images = (cur_input_ids == self.model_config.llm.image_token_index).sum()
            if num_images == 0:
                cur_image_features = image_features[cur_image_idx]
                cur_input_embeds_1 = self.language_model.get_embedding_layer()(cur_input_ids)
                cur_input_embeds = torch.cat([cur_input_embeds_1, cur_image_features[0:0]], dim=0)
                new_input_embeds.append(cur_input_embeds)
                new_labels.append(labels[batch_idx])  # pyright: ignore
                cur_image_idx += 1
                continue

            image_token_indices = (
                [-1]
                + torch.where(cur_input_ids == self.model_config.llm.image_token_index)[0].tolist()
                + [cur_input_ids.shape[0]]
            )
            cur_input_ids_noim = []
            cur_labels = labels[batch_idx]  # pyright: ignore
            cur_labels_noim = []
            for i in range(len(image_token_indices) - 1):
                cur_input_ids_noim.append(
                    cur_input_ids[image_token_indices[i] + 1 : image_token_indices[i + 1]]
                )
                cur_labels_noim.append(
                    cur_labels[image_token_indices[i] + 1 : image_token_indices[i + 1]]
                )
            split_sizes = [x.shape[0] for x in cur_labels_noim]
            cur_input_embeds = self.language_model.get_embedding_layer()(
                torch.cat(cur_input_ids_noim)
            )
            cur_input_embeds_no_im = torch.split(cur_input_embeds, split_sizes, dim=0)
            cur_new_input_embeds = []
            cur_new_labels = []

            for i in range(num_images + 1):
                cur_new_input_embeds.append(cur_input_embeds_no_im[i])
                cur_new_labels.append(cur_labels_noim[i])
                if i < num_images:
                    cur_image_features = image_features[cur_image_idx]
                    cur_image_idx += 1
                    cur_new_input_embeds.append(cur_image_features)
                    cur_new_labels.append(
                        torch.full(
                            (cur_image_features.shape[0],),
                            self.model_config.llm.ignore_index,
                            device=cur_labels.device,
                            dtype=cur_labels.dtype,
                        )
                    )

            cur_new_input_embeds = [x.to(self.device) for x in cur_new_input_embeds]

            cur_new_input_embeds = torch.cat(cur_new_input_embeds)
            cur_new_labels = torch.cat(cur_new_labels)

            new_input_embeds.append(cur_new_input_embeds)
            new_labels.append(cur_new_labels)

        # Truncate sequences to max length as image embeddings can make the sequence longer
        tokenizer_model_max_length = self.language_model.tokenizer.model_max_length
        if tokenizer_model_max_length is not None:
            new_input_embeds = [x[:tokenizer_model_max_length] for x in new_input_embeds]
            new_labels = [x[:tokenizer_model_max_length] for x in new_labels]

        # Combine them
        max_len = max(x.shape[0] for x in new_input_embeds)
        batch_size = len(new_input_embeds)

        new_input_embeds_padded = []
        new_labels_padded = torch.full(
            (batch_size, max_len),
            self.model_config.llm.ignore_index,
            dtype=new_labels[0].dtype,
            device=new_labels[0].device,
        )
        attention_mask = torch.zeros(
            (batch_size, max_len), dtype=attention_mask.dtype, device=attention_mask.device
        )
        position_ids = torch.zeros(
            (batch_size, max_len),
            dtype=position_ids.dtype,  # pyright: ignore
            device=position_ids.device,  # pyright: ignore
        )  # pyright: ignore

        for i, (cur_new_embed, cur_new_labels) in enumerate(
            zip(new_input_embeds, new_labels, strict=False)
        ):
            cur_len = cur_new_embed.shape[0]
            if self.language_model.tokenizer.padding_side == "left":
                new_input_embeds_padded.append(
                    torch.cat(
                        (
                            torch.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                                device=cur_new_embed.device,
                            ),
                            cur_new_embed,
                        ),
                        dim=0,
                    )
                )
                if cur_len > 0:
                    new_labels_padded[i, -cur_len:] = cur_new_labels
                    attention_mask[i, -cur_len:] = True
                    position_ids[i, -cur_len:] = torch.arange(  # pyright: ignore
                        0,
                        cur_len,
                        dtype=position_ids.dtype,  # pyright: ignore
                        device=position_ids.device,  # pyright: ignore
                    )
            else:
                new_input_embeds_padded.append(
                    torch.cat(
                        (
                            cur_new_embed,
                            torch.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                                device=cur_new_embed.device,
                            ),
                        ),
                        dim=0,
                    )
                )
                if cur_len > 0:
                    new_labels_padded[i, :cur_len] = cur_new_labels
                    attention_mask[i, :cur_len] = True
                    position_ids[i, :cur_len] = torch.arange(  # pyright: ignore
                        0,
                        cur_len,
                        dtype=position_ids.dtype,  # pyright: ignore
                        device=position_ids.device,  # pyright: ignore
                    )

        new_input_embeds = torch.stack(new_input_embeds_padded, dim=0)

        if _labels is None:
            new_labels = None
        else:
            new_labels = new_labels_padded

        if _attention_mask is None:
            attention_mask = None
        else:
            attention_mask = attention_mask.to(dtype=_attention_mask.dtype)

        if _position_ids is None:
            position_ids = None

        return None, position_ids, attention_mask, past_key_values, new_input_embeds, new_labels  # pyright: ignore


AutoConfig.register("vlm", VLMConfig)
AutoModel.register(VLMConfig, VLM)



import logging
import os
import random
from pathlib import Path

import hydra
import numpy as np
import torch
from omegaconf import OmegaConf
from transformers import AutoConfig

from .config import AppConfig, ModelConfig, TrainerConfig, register_configs
from .data import get_data_args
from .models import VLM, VLMConfig
from .train import get_training_args, train

log: logging.Logger = logging.getLogger(name=__name__)
CONFIG_PATH: Path = Path(__file__).resolve().parent / "config"


def seed_everything(seed: int) -> None:
    log.info(f"Global seed set to {seed}")
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def print_model(cfg: ModelConfig) -> None:
    components = {
        "model": {"name": cfg.name, "path": CONFIG_PATH / "model" / f"{cfg.name}.yaml"},
        "visual_encoder": {
            "name": cfg.visual_encoder.name,
            "path": CONFIG_PATH / "model" / "visual_encoder" / f"{cfg.visual_encoder.name}.yaml",
        },
        "llm": {
            "name": cfg.llm.name,
            "path": CONFIG_PATH / "model" / "llm" / f"{cfg.llm.name}.yaml",
        },
        "connector": {
            "name": cfg.connector.name,
            "path": CONFIG_PATH / "model" / "connector" / f"{cfg.connector.name}.yaml",
        },
    }

    log.info(
        f"Loading model: [bold red][link=file://{components['model']['path']}]{components['model']['name']}[/link][/bold red]"
    )
    log.info(
        f"Visual encoder: [bold cyan][link=file://{components['visual_encoder']['path']}]{components['visual_encoder']['name']}[/link][/bold cyan]"
    )
    log.info(
        f"LLM: [bold blue][link=file://{components['llm']['path']}]{components['llm']['name']}[/link][/bold blue]"
    )
    log.info(
        f"Connector: [bold yellow][link=file://{components['connector']['path']}]{components['connector']['name']}[/link][/bold yellow]"
    )


def load_model(model_cfg: ModelConfig, trainer_cfg: TrainerConfig, device: torch.device) -> VLM:
    if trainer_cfg.from_pretrained:
        log.info(f"Loading model from pretrained: {trainer_cfg.from_pretrained}")
        model: VLM = VLM.from_pretrained(
            trainer_cfg.from_pretrained,
            torch_device=device,
            attn_implementation=trainer_cfg.attn_implementation,
        )
        return model
    else:
        print_model(model_cfg)
        config: VLMConfig = VLMConfig(
            OmegaConf.to_container(model_cfg, resolve=True),
            "float16" if trainer_cfg.fp16 else ("bfloat16" if trainer_cfg.bf16 else "float32"),
            AutoConfig.from_pretrained(model_cfg.llm.hf_name).hidden_size,
        )
        model = VLM(config, device, trainer_cfg.attn_implementation)
        return model


def vlm(cfg: AppConfig) -> None:
    seed_everything(cfg.trainer.seed)
    if cfg.mode.is_training:
        log.info("Training mode")
        training_args = get_training_args(cfg.trainer)
        model: VLM = load_model(cfg.model, cfg.trainer, training_args.device)
        model.set_trainable_params(cfg.trainer.unfreeze)
        data_args = get_data_args(cfg.dataset, cfg.model, model.visual_encoder.preprocessor)
        train(model, training_args, data_args)


def validate_config(cfg: AppConfig) -> None:
    OmegaConf.to_container(cfg, throw_on_missing=True)


@hydra.main(version_base=None, config_path=str(CONFIG_PATH), config_name="config")  # pyright: ignore
def main(cfg: AppConfig) -> None:
    validate_config(cfg)
    vlm(cfg)


register_configs()

if __name__ == "__main__":
    main()



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, cast, override

import torch.nn as nn
from torch import FloatTensor, LongTensor, Tensor, device, dtype
from transformers import (
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from transformers.utils import ModelOutput

from ...config.config_schema import LLMConfig

log: logging.Logger = logging.getLogger(name=__name__)


@dataclass
class TokenConfig:
    """Special token configuration"""

    image_token: str = "<image>"
    image_patch_token: str = "<im_patch>"
    image_start_token: str = "<im_start>"
    image_end_token: str = "<im_end>"
    system_token: str | None = None
    user_token: str | None = None
    assistant_token: str = "<|assistant|>"
    image_token_id: int | None = None
    image_patch_token_id: int | None = None
    image_start_token_id: int | None = None
    image_end_token_id: int | None = None


@dataclass
class LanguageModelConfig:
    hidden_size: int | None = None
    vocab_size: int | None = None
    max_seq_length: int | None = None


class LanguageModel(nn.Module, ABC):
    def __init__(
        self,
        config: LLMConfig,
        torch_dtype: dtype,
        torch_device: device,
        attn_implementation: str = "eager",
    ) -> None:
        super().__init__()
        self.config: LLMConfig = config
        self.name: str = self.config.name
        self.hf_name: str = self.config.hf_name
        self.model_type: str = self.config.type
        self.torch_dtype: dtype = torch_dtype
        self.torch_device: device = torch_device
        self.attn_implementation: str = attn_implementation

        # model config
        self.model_config: LanguageModelConfig = LanguageModelConfig(
            hidden_size=getattr(self.config, "hidden_size", None),
            vocab_size=getattr(self.config, "vocab_size", None),
            max_seq_length=getattr(self.config, "max_seq_length", None),
        )

        # token config
        self.token_config: TokenConfig = TokenConfig(
            image_token=getattr(self.config, "image_token", "<image>"),
            image_patch_token=getattr(self.config, "image_patch_token", "<im_patch>"),
            image_start_token=getattr(self.config, "image_start_token", "<im_start>"),
            image_end_token=getattr(self.config, "image_end_token", "<im_end>"),
            system_token=getattr(self.config, "system_token", None),
            user_token=getattr(self.config, "user_token", None),
            assistant_token=getattr(self.config, "assistant_token", "<|assistant|>"),
        )

        self.initialize_components()

    # initialize all components
    def initialize_components(self) -> None:
        self._hf_config: PretrainedConfig = self._build_hf_config()
        self.verify_config()
        self._tokenizer: PreTrainedTokenizer = self._build_tokenizer()
        self._tokenizer.pad_token = self._tokenizer.unk_token
        self._add_special_tokens()
        self._language_model: PreTrainedModel = self._build_language_model()
        self.language_model.resize_token_embeddings(len(self.tokenizer))

    def _add_special_tokens(self) -> None:
        """Adds special tokens to the tokenizer if they don't exist."""
        # Create a mapping of tokens to their attribute names

        token_mapping = {
            self.token_config.system_token: "system_token_id",
            self.token_config.user_token: "user_token_id",
            self.token_config.assistant_token: "assistant_token_id",
        }
        if self.config.use_image_patch_token:
            token_mapping[self.token_config.image_patch_token] = "image_patch_token_id"
        if self.config.use_start_end_tokens:
            token_mapping[self.token_config.image_start_token] = "image_start_token_id"
            token_mapping[self.token_config.image_end_token] = "image_end_token_id"

        # Identify which tokens need to be added
        tokens_to_add: list[Any] = []
        for token in token_mapping:
            if token is None:
                continue
            if self.tokenizer.convert_tokens_to_ids(token) == self.tokenizer.unk_token_id:
                tokens_to_add.append(token)
                log.info(f"Token '{token}' does not exist in tokenizer, will be added")
            else:
                token_id = self.tokenizer.convert_tokens_to_ids(token)
                log.info(f"Token '{token}' exists in tokenizer, ID: {token_id}")
                setattr(self.token_config, token_mapping[token], cast(int, token_id))

        # Add all new tokens at once if any
        if tokens_to_add:
            log.info(f"Adding tokens: {tokens_to_add}")
            self.tokenizer.add_tokens(tokens_to_add, special_tokens=True)

            # Now set the IDs for newly added tokens
            for token in tokens_to_add:
                token_id = cast(int, self.tokenizer.convert_tokens_to_ids(token))
                setattr(self.token_config, token_mapping[token], token_id)

    @property
    def tokenizer(self) -> PreTrainedTokenizer:
        return self._tokenizer

    @property
    def language_model(self) -> PreTrainedModel:
        return self._language_model

    @property
    def hf_config(self) -> PretrainedConfig:
        return self._hf_config

    @property
    def hidden_size(self) -> int:
        return cast(int, self.model_config.hidden_size)

    @hidden_size.setter
    def hidden_size(self, value: int) -> None:
        self.model_config.hidden_size = value

    @property
    def vocab_size(self) -> int:
        return cast(int, self.model_config.vocab_size)

    @vocab_size.setter
    def vocab_size(self, value: int) -> None:
        self.model_config.vocab_size = value

    @property
    def max_seq_length(self) -> int:
        return cast(int, self.model_config.max_seq_length)

    @max_seq_length.setter
    def max_seq_length(self, value: int) -> None:
        self.model_config.max_seq_length = value

    @property
    def image_token_id(self) -> int:
        return cast(int, self.token_config.image_token_id)

    @property
    def pad_token_id(self) -> int:
        return cast(int, self.tokenizer.pad_token_id)

    @abstractmethod
    def get_embedding_layer(self) -> nn.Module:
        pass

    @abstractmethod
    def _build_tokenizer(self) -> PreTrainedTokenizer:
        pass

    @abstractmethod
    def _build_language_model(self) -> PreTrainedModel:
        pass

    @abstractmethod
    def _build_hf_config(self) -> PretrainedConfig:
        pass

    @override
    def forward(
        self,
        input_ids: Tensor | None = None,
        inputs_embeds: Tensor | None = None,
        attention_mask: Tensor | None = None,
        position_ids: LongTensor | None = None,
        past_key_values: list[FloatTensor] | None = None,
        labels: LongTensor | None = None,
        use_cache: bool | None = None,
        output_attentions: bool | None = None,
        output_hidden_states: bool | None = None,
        images: FloatTensor | None = None,
        image_sizes: list[list[int]] | None = None,
        return_dict: bool | None = None,
    ) -> Tensor:
        return self.language_model(
            input_ids=input_ids,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            labels=labels,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

    @abstractmethod
    def generate(
        self,
        position_ids: LongTensor | None = None,
        attention_mask: LongTensor | None = None,
        inputs_embeds: FloatTensor | None = None,
        **kwargs: Any,
    ) -> ModelOutput | LongTensor:
        pass

    def verify_config(self) -> None:
        config_pairs = [
            ("hidden_size", self.get_config("hidden_size"), self.hidden_size),
            ("vocab_size", self.get_config("vocab_size"), self.vocab_size),
            ("max_seq_length", self.get_config("max_position_embeddings"), self.max_seq_length),
        ]

        for key, model_value, config_value in config_pairs:
            self._verify_param_match(key, model_value, config_value)

    def _verify_param_match(
        self, key: str, model_value: int | str | None, config_value: int | str | None
    ) -> None:
        capitalized_key = key.capitalize()

        if model_value is None and config_value is None:
            log.warning(f"{capitalized_key} not found in config for {self.name}")
        elif model_value is not None and config_value is None:
            setattr(self, key, int(model_value))
            if hasattr(self.config, key):
                setattr(self.config, key, int(model_value))
            log.info(f"{capitalized_key} not found in config, using hf config: {model_value}")
        elif model_value is None and config_value is not None:
            log.warning(f"{capitalized_key} not found in hf config for {self.name}")
        elif model_value is not None and config_value is not None:
            if model_value != config_value:
                error_msg = f"{capitalized_key} mismatch: hf config: {model_value} != config: {config_value}"
                log.warning(error_msg)
            else:
                log.info(
                    f"{capitalized_key} verified: hf config: {model_value} == config: {config_value}"
                )

    def get_config(self, key: str) -> int | str | None:
        return getattr(self.hf_config, key, None)



from .inference import process_images, tokenizer_image_token
from .models import VLM, VLMConfig
from .utils import conversation

__all__ = [
    "VLM",
    "VLMConfig",
    "process_images",
    "tokenizer_image_token",
    "conversation",
]
